/*
 * This file is part of mpv.
 *
 * mpv is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * mpv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with mpv.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MPV_MACOSX_VERSIONS
#define MPV_MACOSX_VERSIONS

#if !defined(MAC_OS_X_VERSION_10_9)
#    define MAC_OS_X_VERSION_10_9 1090
#endif

#if !defined(MAC_OS_X_VERSION_10_10)
#    define MAC_OS_X_VERSION_10_10 101000
#endif

#if !defined(MAC_OS_X_VERSION_10_12)
#    define MAC_OS_X_VERSION_10_12 101200
#endif

#endif /* MPV_MACOSX_VERSIONS */
